# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import fcntl
import logging
import os
import signal
import subprocess
import time

from cmf.monitor import schema
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class SubprocessTimeout(object):
  """
  SubprocessTimeout is an utility class that supports
  running of a command in a subprocess with a timeout.
  """
  def _non_block_read(self, output):
    """
    Sets a file descriptor to non-blocking mode and 
    returns the result of a read from the stream. 
    Restores the descriptor to its original status
    """
    fd = output.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
    try:
      try:
        return output.read()
      except:
        return ''
    finally:
      fcntl.fcntl(fd, fcntl.F_SETFL, fl)

  def subprocess_with_timeout(self, args, timeout):
    """
    Execute args in a subprocess.
    Limit execution time to 'timeout' seconds.
    """
    p = subprocess.Popen(args,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         close_fds=True)
    t_begin = time.time()
    seconds_passed = 0
    stdout = ''
    stderr = ''

    while p.poll() is None and seconds_passed < timeout:
      # p.std* blocks on read(), which messes up the timeout timer.
      stdout += self._non_block_read(p.stdout)
      stderr += self._non_block_read(p.stderr)
      time.sleep(0.1)
      seconds_passed = time.time() - t_begin
      
    if seconds_passed >= timeout:
      try:
        p.stdout.close()
        p.stderr.close()
        # use os.kill instead of p.terminate for python 2.4 compatibility
        os.kill(p.pid, signal.SIGTERM)
      except:
        THROTTLED_LOG.exception("Kill subprocess exception with args %s" % args)
        raise Exception("failed timeout with args %s" % args)

      THROTTLED_LOG.exception("Timeout with args %s" % args)
      raise Exception("timeout with args %s" % args)
    
    stdout += self._non_block_read(p.stdout)
    stderr += self._non_block_read(p.stderr)
    return (p.returncode, stdout, stderr)
