// (c) Copyright 2013 Cloudera, Inc. All rights reserved.
define([
  'cloudera/Util',
  'underscore'
], function(Util, _) {

  var assert = function(result, message) {
    if (!result) {
      throw new Error(message);
    }
  };

  // Returns a function to be used as the dataSource property for a
  // SearchResults object.
  // 
  // Options:
  // * executeQuery (required) Function that takes an object defining
  //   the response handlers and makes the actual query to the server.
  // * getResults (required) Function that returns list of results given
  //   the actual server response.
  // * orderingProperty (required) String property of individual result
  //   to order the results by.
  // * getMaximum (required) Function that returns maximum possible value
  //   of orderingProperty.
  // * onDataReceived (optional) Called with data from the server once
  //   response comes back.
  return function(options) {
    // Quickly verify that we're given the functions we need.
    assert(_.isFunction(options.executeQuery),
      'executeQuery must be present and a function');
    assert(_.isFunction(options.getResults),
      'getResults must be present and a function');
    assert(options.orderingProperty, 'orderingProperty must be present');
    assert(_.isFunction(options.getMaximum),
      'getMaximum must be present and a function');

    // Monotonically incrementing number that helps ensure we don't send
    // multiple inflight requests.
    var state = 0;

    // Return a function suitable for use in the later call to retrieve
    // history events from the server.
    var createOnRetrieveHistorySuccess = function(searchResults, clear) {
      // Persist the state at the time of the request.
      var requestState = state, 
      // Persist the length of the events at the time on the request.
      initialLength = clear ? 0 : searchResults.length();

      return function(data) {
        if (data && _.isFunction(options.onDataReceived)) {
          options.onDataReceived(data);
        }

        var currentLength = searchResults.length(),
        searcher, newest, oldest, prependRange;
        // Ignore the result if the current state doesn't equal 
        // the state when the request was made
        if (requestState !== state
            // or there's no data returned
            || data === null
            // or another request beat this request in.
            || initialLength !== currentLength) {
          return;
        }
        var results = options.getResults(data);
        // Check if any data was returned from options.getResults.
        if (_.isEmpty(results)) {
          return;
        }

        // If the list is currently empty, just push the results
        // into the list.
        if (!currentLength) {
          searchResults.append(results);
          return;
        }

        var orderingProperty = options.orderingProperty;

        oldest = searchResults.getTail()[orderingProperty];
        // Check the ordering property of our "oldest" record. If it is newer
        // then append the result.
        if (results[0][orderingProperty] < oldest) {
          searchResults.append(results);
        } else {
          // Otherwise, prepend the results.
          results.reverse();
          searcher = function(i) {
            return results[i][orderingProperty];
          };

          newest = searchResults.getHead()[orderingProperty];
          var maximum = options.getMaximum();
          prependRange = Util.findRange(searcher, results.length-1, newest+1, maximum);
          if (prependRange) {
            searchResults.prepend(results.slice(prependRange.start, prependRange.end).reverse());
          }
        }
      };
    };

    var dataSource = function(searchResults, clear, callback) {
      state += 1;
      // Stores any error that occurs during the request phase.
      var err;
      // Clear the results if they're no longer applicable.
      if (clear) {
        searchResults.clear();
      }
      // Create the response handlers.
      var handlers = {
        success: createOnRetrieveHistorySuccess(searchResults, clear),
        error: function(jqXHR, textStatus, errorThrown) {
            // Store the error for later use.
            err = errorThrown;
        },
        complete: function() {
          // Invoke the callback, passing an error if one occurred.
          callback(err);
        }
      };
      // Execute the query.
      options.executeQuery(handlers);
    };
    // Expose the createOnRetrieveHistorySuccess function.
    dataSource.createOnRetrieveHistorySuccess = createOnRetrieveHistorySuccess;
    return dataSource;
  };
});